<?php

use Aws\Common\Aws;
use Aws\Sns\Message;
use Aws\Sns\MessageValidator;
use TNP\Mailer\PHPMailerLoader;

class NewsletterAmazon extends NewsletterMailerAddon {

    /**
     * @var NewsletterAmazon
     */
    static $instance;
    var $ses_client = null;
    var $notification_endpoint = null;

    /* @var NewsletterLogger */
    var $snslogger = null;

    function __construct($version = '0.0.0') {
        self::$instance = $this;
        parent::__construct('amazon', $version);
        $this->notification_endpoint = $this->get_bounce_endpoint();
        // To init the internal $this->logger used in this plugin (to be fixed)
        $this->get_logger();
        $this->snslogger = new NewsletterLogger('amazon-bounce');
    }

    function get_description() {
        return 'Amazon Addon';
    }

    function get_mailer() {
        static $mailer = null;
        if (!$mailer) {
            $mailer = new NewsletterAmazonMailer($this);
        }
        return $mailer;
    }

    function init() {

        parent::init();

        // catch bounce action
        add_action('newsletter_action', array($this, 'hook_tnp_bounce'));

        if (is_admin() && current_user_can('administrator')) {

            add_action('admin_menu', array($this, 'hook_admin_menu'), 100);
            add_action('wp_ajax_tnp_amazon_checkaddr', array($this, "hook_wp_ajax_tnp_amazon_checkaddr"));
            add_action('wp_ajax_tnp_amazon_verifyaddr', array($this, "hook_wp_ajax_tnp_amazon_verifyaddr"));
            add_action('wp_ajax_tnp_amazon_bouncecheck', array($this, "hook_wp_ajax_tnp_amazon_bouncecheck"));

            add_filter('newsletter_menu_settings', array($this, 'hook_newsletter_menu_settings'));
        }
    }

    /**
     * Handle bounce action
     * @param $action
     */
    function hook_tnp_bounce($action) {

        if ($action == 'bounce') {

            $this->snslogger->debug('Handle bounce action');

            // Make sure the request is POST
            if ($_SERVER['REQUEST_METHOD'] !== 'POST') {
                $this->snslogger->debug('405 Method not allowed');
                http_response_code(405);
                die;
            }

            $this->loadAws();

            try {

                // Create a message object from the POST body
                $message = Message::fromRawPostData();
                $this->snslogger->debug($message);

                // Validate the message
                $validator = new MessageValidator();
                $validator->validate($message);
                $type = $message['Type'];
                $this->snslogger->debug('--- Handle the [' . $type . '] notification');

                // handle the notification
                if ($type === 'SubscriptionConfirmation') {
                    $this->snslogger->debug('Type: SubscriptionConfirmation');
                    $this->confirmSubscription($message['SubscribeURL']);
                } elseif ($type === 'Notification') {
                    $this->snslogger->debug('Type: Notification');
                    $this->processBounce($message['Message']);
                }
            } catch (Exception $ex) {
                // handle errors
                $this->snslogger->debug($ex->getMessage());
                http_response_code(404);
                die();
            }
        }
    }

    function hook_newsletter_menu_settings($entries) {
        $entries[] = array('label' => '<i class="fas fa-envelope"></i> Amazon SES', 'url' => '?page=newsletter_amazon_index', 'description' => 'Send emails with Amazon SES');
        return $entries;
    }

    function hook_admin_menu() {
        add_submenu_page('newsletter_main_index', 'Amazon SES Extension', '<span class="tnp-side-menu">Amazon SES</span>', 'manage_options', 'newsletter_amazon_index', array($this, 'menu_page_index'));
    }

    function menu_page_index() {
        global $wpdb;
        require dirname(__FILE__) . '/index.php';
    }

    function hook_wp_ajax_tnp_amazon_checkaddr() {

        if (!wp_verify_nonce($_REQUEST['nonce'], "newsletter_amazon_checkaddr_nonce")) {
            exit("No naughty business please");
        }

        if (!current_user_can('administrator')) {
            exit("Not enough privileges");
        }

        $address = Newsletter::instance()->options['sender_email'];
        $ajax_result['valid'] = -1;
        $ajax_result['message'] = "";

        try {

            if (defined('NEWSLETTER_AMAZON_KEY')) {
                $ses_client = $this->getSesClient(NEWSLETTER_AMAZON_KEY, NEWSLETTER_AMAZON_SECRET, $_POST['aws_region']);
            } else {
                $ses_client = $this->getSesClient($_POST['aws_key'], $_POST['aws_secret'], $_POST['aws_region']);
            }

            if (is_null($ses_client)) {
                $ajax_result['message'] = 'Error creating SES client, check AWS credentials provided.';
            } else {
                // get the verification status of the identity
                $this->logger->debug("Calling getIdentityVerificationAttributes");

                $domain = substr($address, strrpos($address, '@') + 1);

                $result = $ses_client->getIdentityVerificationAttributes(array(
                    'Identities' => array($address, $domain),
                ));

                $this->logger->debug($result);

                $attributes = $result->get('VerificationAttributes');

                if (is_array($attributes[$address])) {
                    if ($attributes[$address]['VerificationStatus'] == 'Success') {
                        $ajax_result['valid'] = 1;
                        $ajax_result['message'] = 'Address ' . $address . ' <strong>verified</strong>.';
                    } else {
                        $ajax_result['valid'] = 0;
                        $ajax_result['message'] = '<strong>Pending verification</strong>, check your inbox.';
                    }
                } elseif (is_array($attributes[$domain])) {
                    if ($attributes[$domain]['VerificationStatus'] == 'Success') {
                        $ajax_result['valid'] = 1;
                        $ajax_result['message'] = 'Domain ' . $domain . ' <strong>verified</strong>.';
                    }
                } else {
                    $ajax_result['valid'] = 0;
                    $ajax_result['message'] = 'Address or domain ' . $address . ' <strong>NOT verified<strong>.';
                }
            }
        } catch (Exception $e) {
            $m = $e->getMessage();
            if (stripos($m, 'AccessDenied') !== false) {
                $ajax_result['message'] = '<span style="color: red">The API key and secret seem associated to a Amazon AWS user (IAM) without enough permissions.</span> '
                        . '<a href="https://www.thenewsletterplugin.com/documentation/addons/delivery-addons/amazon-ses-extension/#iam-user" target="_blank">Read here to solve this problem</a>.';
            } else {
                $ajax_result['message'] = '<span style="color: red">Error checking verification status, please check you credentials and region.</span>';
            }

            $ajax_result['message'] .= '<br><pre style="font-size: .9em">' . esc_html($e->getMessage()) . '</pre>';

            //$ajax_result['message'] = '<span style="color: red">Error checking verification status, please check you credentials and region.</span>';
            $this->logger->debug($e->getMessage());
            $this->logger->debug($e->getAwsErrorMessage());
        }

        if (!empty($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) == 'xmlhttprequest') {
            echo json_encode($ajax_result);
        } else {
            header("Location: " . $_SERVER["HTTP_REFERER"]);
        }

        wp_die();
    }

    function hook_wp_ajax_tnp_amazon_verifyaddr() {

        if (!wp_verify_nonce($_REQUEST['nonce'], "newsletter_amazon_verifyaddr_nonce")) {
            exit("No naughty business please");
        }

        if (!current_user_can('administrator')) {
            exit("Not enough privileges");
        }

        $address = Newsletter::instance()->options['sender_email'];
        $ajax_result['valid'] = -1;
        $ajax_result['message'] = "";

        try {

            if (defined('NEWSLETTER_AMAZON_KEY')) {
                $ses_client = $this->getSesClient(NEWSLETTER_AMAZON_KEY, NEWSLETTER_AMAZON_SECRET, $_POST['aws_region']);
            } else {
                $ses_client = $this->getSesClient($_POST['aws_key'], $_POST['aws_secret'], $_POST['aws_region']);
            }

            if (is_null($ses_client)) {
                $ajax_result['message'] = 'Error creating SES client, check AWS credentials provided.';
            } else {

                // verify the identity
                $result = $ses_client->verifyEmailIdentity(array(
                    'EmailAddress' => $address,
                ));

                $ajax_result['valid'] = 1;
                $ajax_result['message'] = 'Email address verification requested. This action causes a <strong>confirmation email message</strong> to be sent to the specified address. Now <strong>check your inbox</strong>.';
            }
        } catch (Exception $e) {
            $m = $e->getMessage();
            if (stripos($m, 'AccessDenied') !== false) {
                $ajax_result['message'] = '<span style="color: red">The API key and secret seem associated to a Amazon AWS user (IAM) without enough permissions.</span> '
                        . '<a href="https://www.thenewsletterplugin.com/documentation/addons/delivery-addons/amazon-ses-extension/#iam-user" target="_blank">Read here to solve this problem</a>.'
                        . '<br><pre style="font-size: .9em">' . esc_html($e->getMessage()) . '</pre>';
            } else {
                $ajax_result['message'] = '<pre style="font-size: .9em">' . esc_html($e->getMessage()) . '</pre>';
            }
        }

        if (!empty($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) == 'xmlhttprequest') {
            echo json_encode($ajax_result);
        } else {
            header("Location: " . $_SERVER["HTTP_REFERER"]);
        }

        wp_die();
    }

    function hook_wp_ajax_tnp_amazon_bouncecheck() {

        if (!wp_verify_nonce($_REQUEST['nonce'], "newsletter_amazon_bouncecheck_nonce")) {
            exit("No naughty business please");
        }

        if (!current_user_can('administrator')) {
            exit("Not enough privileges");
        }

        // Prepare default ajax result
        $ajax_result['topicARN'] = '-';
        $ajax_result['topicSubscriptions'] = '-';
        $ajax_result['bounceDestination'] = '-';
        $ajax_result['complaintDestination'] = '-';

        try {

            $sns_client = $this->getSnsClient();
            $ses_client = $this->getSesClient();

            if (is_null($ses_client)) {

                $ajax_result['error'] = 'Error creating SES or SNS client, check AWS credentials provided.';
            } else {

                // check bounce notification settings
                $address = Newsletter::instance()->options['sender_email'];
                $domain = substr($address, strrpos($address, '@') + 1);

                $notificationAttributes = $ses_client->GetIdentityNotificationAttributes(array(
                            'Identities' => array($address, $domain),
                        ))->get('NotificationAttributes');
                $this->logger->debug('SES Notification Attributes');
                $this->logger->debug($notificationAttributes);
                foreach ($notificationAttributes as $notificationAttribute) {

                    if (!empty($notificationAttribute['BounceTopic'])) {
                        $bounce_array[] = $notificationAttribute['BounceTopic'];
                    }

                    if (!empty($notificationAttribute['ComplaintTopic'])) {
                        $complaint_array[] = $notificationAttribute['ComplaintTopic'];
                    }

                    if (!empty($notificationAttribute['ForwardingEnabled'])) {
                        $bounce_array[] = 'Email Feedback Forwarding';
                    }
                }

                if (!empty($bounce_array)) {
                    $ajax_result['bounceDestination'] = implode(', ', $bounce_array);
                } else {
                    $ajax_result['bounceDestination'] = 'Bounce destination not set for sender address or domain';
                }

                if (!empty($bounce_array)) {
                    $ajax_result['complaintDestination'] = implode(', ', $complaint_array);
                } else {
                    $ajax_result['complaintDestination'] = 'Complaint destination not set for sender address or domain';
                }

                $topic_ARN = $this->options['bounce_topic_arn'];
                $topic_list = $sns_client->listTopics()->get('Topics');

                $this->logger->debug('SNS Topics list');
                $this->logger->debug($topic_list);

                // check if saved or old topic exists
                if ($topic_ARN = $this->is_topic_in_sns_topic_list($topic_ARN, $topic_list)) {

                    $ajax_result['topicARN'] = $topic_ARN;

                    $subscription_list = $sns_client->listSubscriptionsByTopic(array('TopicArn' => $topic_ARN))->get('Subscriptions');
                    $this->logger->debug($subscription_list);
                    foreach ($subscription_list AS $sub) {
                        $subscriptions_array[] = $sub['Endpoint'];
                    }

                    if (!empty($subscriptions_array)) {
                        $ajax_result['topicSubscriptions'] = implode(', ', $subscriptions_array);
                    } else {
                        $ajax_result['topicSubscriptions'] = 'No subscriptions found, check them on SNS console. You have to subscribe the URL <em>' . $this->get_bounce_endpoint() . '</em>';
                    }
                }
            }
        } catch (Exception $e) {
            $m = $e->getMessage();
            if (stripos($m, 'AccessDenied') !== false) {
                $ajax_result['error'] = '<span style="color: red">The API key and secret seem associated to a Amazon AWS user (IAM) without enough permissions.</span> '
                        . '<a href="https://www.thenewsletterplugin.com/documentation/addons/delivery-addons/amazon-ses-extension/#iam-user" target="_blank">Read here to solve this problem</a>.'
                        . '<br><pre style="font-size: .9em">' . esc_html($e->getMessage()) . '</pre>';
            } else {
                $ajax_result['error'] = '<pre style="font-size: .9em">' . esc_html($e->getMessage()) . '</pre>';
            }
        }

        if (!empty($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) == 'xmlhttprequest') {
            echo json_encode($ajax_result);
        } else {
            header("Location: " . $_SERVER["HTTP_REFERER"]);
        }

        wp_die();
    }

    function hook_wp_ajax_tnp_must_login() {
        echo "You must log in";
        wp_die();
    }

    var $amazon = null;
    var $amazon_result;

    function loadAws() {
        if (!class_exists('\Aws\Sdk')) {
            $this->logger->debug('AWS loaded by us');
            require_once __DIR__ . '/api/aws-autoloader.php';
        } else {
            $this->logger->debug('AWS already loaded:' . \Aws\Sdk::VERSION);
        }
    }

    /**
     * @param string $aws_key
     * @param string $aws_secret
     * @param string $aws_region
     * @return \Aws\Ses\SesClient
     */
    function getSesClient($aws_key = null, $aws_secret = null, $aws_region = null) {

        $this->loadAws();

        if (defined('NEWSLETTER_AMAZON_KEY')) {
            $aws_key = NEWSLETTER_AMAZON_KEY;
            $aws_secret = NEWSLETTER_AMAZON_SECRET;
        } else {

            if (empty($aws_key)) {
                $aws_key = $this->options['aws_key'];
            }
            if (empty($aws_secret)) {
                $aws_secret = $this->options['aws_secret'];
            }
        }
        if (empty($aws_region)) {
            $aws_region = $this->options['aws_region'];
        }

        $this->logger->debug('Creating SES client');

        return new \Aws\Ses\SesClient([
            'credentials' => [
                'key' => $aws_key,
                'secret' => $aws_secret
            ],
            'region' => $aws_region,
            'version' => '2010-12-01',
            'ssl.certificate_authority' => false
        ]);
    }

    function getSnsClient() {

        $this->loadAws();
        if (defined('NEWSLETTER_AMAZON_KEY')) {
            $aws_key = NEWSLETTER_AMAZON_KEY;
            $aws_secret = NEWSLETTER_AMAZON_SECRET;
        } else {
            $aws_key = $this->options['aws_key'];
            $aws_secret = $this->options['aws_secret'];
        }
        $aws_region = $this->options['aws_region'];

        $this->logger->debug('Creating SNS client');

        return new \Aws\Sns\SnsClient([
            'credentials' => [
                'key' => $aws_key,
                'secret' => $aws_secret
            ],
            'region' => $aws_region,
            'version' => '2010-03-31',
            'ssl.certificate_authority' => false
        ]);
    }

    function generate_topic_name() {
        $prefix = 'tnp_bounce_';
        return $prefix . sanitize_title(preg_replace("(^https?://)", "", home_url()));
    }

    function get_bounce_endpoint() {
        $homeUrl = home_url();
        if (substr($homeUrl, -1) !== '/')
            $homeUrl .= '/';
        return $homeUrl . '?na=bounce';
    }

    function is_topic_in_sns_topic_list($topic_ARN, $sns_topic_list) {
        foreach ($sns_topic_list as $topic) {
            if ($topic_ARN === $topic['TopicArn']) {
                return $topic_ARN;
            } else if (stripos($topic['TopicArn'], 'wp_newsletter_bounce')) {
                return $topic['TopicArn'];
            }
        }
        return false;
    }

    function is_notification_endpoint_exist_in_sns_subscription_list($sns_subscription_list) {
        foreach ($sns_subscription_list as $subscription) {
            if ($this->notification_endpoint === $subscription['Endpoint']) {
                return true;
            }
        }

        return false;
    }

    function get_topic_name_from_topic_ARN($topic_ARN) {
        if (is_string($topic_ARN) && strpos($topic_ARN, ':') !== false) {
            return end(explode(':', $topic_ARN));
        }

        return null;
    }

    /*
     * send a request to the SubscribeURL to complete subscription
     */

    function confirmSubscription($url) {
        $this->snslogger->debug("SubscriptionConfirmation [URL: $url]");
        wp_remote_get($url);
    }

    /*
     * handle the bounce
     */

    function processBounce($message) {

        global $wpdb;

        $this->snslogger->debug("Handle the message");
        // update last run
        update_option('newsletter_amazon_last_run', time());

        // check if we have to process soft bounces and complaints
        $process_soft_bounces = $this->options['process_soft_bounces'] ? true : false;
        $this->snslogger->debug('Process soft bounces: ' . $process_soft_bounces);

        $this->snslogger->debug($message);

        $message = json_decode($message);

        $recipients = null;

        if ($message->notificationType == "Bounce") {

            $bounceType = $message->bounce->bounceType;
            //$this->snslogger->debug('Bounce type: ' . $bounceType);

            if ($bounceType == 'Permanent' || $process_soft_bounces) {

                $recipients = $message->bounce->bouncedRecipients;
                if (is_array($recipients) || is_object($recipients)) {
                    //$this->snslogger->debug('[' . count($recipients) . '] recipients');

                    $query = "update " . NEWSLETTER_USERS_TABLE . " set status='B' where email=%s limit 1";

                    foreach ($recipients as $recipient) {
                        $email = $recipient->emailAddress;
                        $this->snslogger->debug('Bounced address: ' . $this->obfuscate_email($email));
                        $wpdb->query($wpdb->prepare($query, strtolower($email)));
                    }
                }
            }
        }


        if ($message->notificationType == "Complaint") {
            //$this->snslogger->debug("Complaint");
            $recipients = $message->complaint->complainedRecipients;
            if (is_array($recipients) || is_object($recipients)) {
                $query = "update " . NEWSLETTER_USERS_TABLE . " set status='P' where email=%s limit 1";
                foreach ($recipients as $recipient) {
                    $email = $recipient->emailAddress;
                    $this->snslogger->debug('Complained address: ' . $this->obfuscate_email($email));
                    $wpdb->query($wpdb->prepare($query, strtolower($email)));
                }
            }
        }
    }

    function obfuscate_email($email) {
        $em = explode("@", $email);
        $name = implode('@', array_slice($em, 0, count($em) - 1));
        $len = floor(strlen($name) / 2);
        return substr($name, 0, $len) . str_repeat('*', $len) . "@" . end($em);
    }

}

class NewsletterAmazonMailer extends NewsletterMailer {

    var $client;

    /**
     * @var PHPMailer 
     */
    var $phpMailer;

    /**
     * @var NewsletterAmazon
     */
    var $addon;

    /**
     *
     * @param NewsletterAmazon $addon
     */
    function __construct($addon) {
        parent::__construct($addon->name, $addon->options);
        $this->addon = $addon;
    }

    public function get_description() {
        return 'Amazon SES Addon';
    }

    /**
     * 
     * @return \Aws\Ses\SesClient
     */
    function getSesClient() {

        // If the default SES client is request and we have it, return it
        if (empty($this->client)) {
            $this->client = $this->addon->getSesClient();
        }

        return $this->client;
    }

    function getPhpMailer() {
        if (!$this->phpMailer) {
            require_once NEWSLETTER_INCLUDES_DIR . '/PHPMailerLoader.php';
            $this->phpMailer = PHPMailerLoader::make_instance();
            $this->phpMailer->CharSet = 'UTF-8';
            PHPMailer::$validator = 'noregex';
        }

        $this->phpMailer->clearAllRecipients();
        $this->phpMailer->clearCustomHeaders();

        return $this->phpMailer;
    }

    /**
     * @param TNP_Mailer_Message $message
     * @return array[]
     * @throws phpmailerException
     */
    function prepareEmail($message) {

        $newsletter = Newsletter::instance();

        $mailer = $this->getPhpMailer();

        if (!empty($message->body_text)) {
            $text = $message->body_text;
        }

        $mailer->setFrom($message->from, $message->from_name);

        $reply_to = $newsletter->options['reply_to'];
        if (!empty($reply_to)) {
            $mailer->addReplyTo($reply_to);
        }

        $return_path = $newsletter->options['return_path'];
        if (!empty($return_path)) {
            $mailer->Sender = $return_path;
        }

        foreach ($message->headers AS $name => $value) {
            $mailer->addCustomHeader($name, $value);
        }

        $mailer->addAddress($message->to);
        $mailer->Subject = $message->subject;
        $mailer->Body = $message->body;
        $mailer->AltBody = $message->body_text;

        $mailer->IsHTML(true);

        $mailer->preSend();

        return array(
            'RawMessage' => array(
                'Data' => $mailer->getSentMIMEMessage()
            )
        );
    }

    /**
     *
     * @param TNP_Mailer_Message $message
     * @return WP_Error|boolean
     */
    function send($message) {

        $logger = $this->get_logger();

        try {

            $logger->debug('Sending to: ' . NewsletterAmazon::$instance->obfuscate_email($message->to));
            $ses_client = $this->getSesClient();

            $ses_email = $this->prepareEmail($message);

            $result = $ses_client->sendRawEmail($ses_email);

            //$logger->debug('Sent to: ' .  NewsletterAmazon::$instance->obfuscate_email($message->to));
            // Look there at the page end
            // https://docs.aws.amazon.com/sdk-for-php/v3/developer-guide/guide_commands.html
            gc_collect_cycles();
        } catch (Exception $e) {
            $logger->error($e->getMessage());
            return new WP_Error(self::ERROR_GENERIC, $e->getMessage());
        }
    }

    public function send_chunk($messages) {

        $logger = $this->get_logger();

        $ses_client = $this->getSesClient();

        $commands = array();

        foreach ($messages as $message) {

            $ses_email = $this->prepareEmail($message);
            $logger->debug('Sending to: ' . NewsletterAmazon::$instance->obfuscate_email($message->to));
            $commands[] = $ses_client->getCommand('SendRawEmail', $ses_email);
        }

        try {

            // TODO: Manage the promise results
            $pool = new \Aws\CommandPool($ses_client, $commands);
            // Initiate the pool transfers
            $promise = $pool->promise();

            // Force the pool to complete synchronously
            $promise->wait();

            // Look there at the page end
            // https://docs.aws.amazon.com/sdk-for-php/v3/developer-guide/guide_commands.html            
            gc_collect_cycles();
        } catch (Exception $e) {
            $logger->error($e->getMessage());
            return new WP_Error(self::ERROR_GENERIC, $e->getMessage());
        }

        return true;
    }

}
